'' =================================================================================================
''
''   File....... jm_wav_player.spin
''   Purpose.... Mono/Stereo WAV player object for 16-bit mono or stereo WAV files
''   Author..... Jon "JonnyMac" McPhalen
''               -- Copyright (c) 2010-2013 Jon McPhalen
''               -- based on original code by Ray Allen
''               -- see below for terms of use
''   E-mail..... teamefx@efx-tek.com
''   Started.... 
''   Updated.... 19 MAY 2013
''
'' =================================================================================================

{

  Audio player object for 16-bit mono or stereo WAV files. Files must be in canonical WAV format,
  i.e., without metadata. Use Audacity to strip non-audio data from files.

  Features:
    -- playback of mono or stereo WAV files
    -- output to one (mono only) or two (mono and stereo) channels
    -- volume level indepent by channel; may be set while file plays
    -- playback rate (50% to 200%) may be modified while file is playing
       * maximum playback rate is 48kHz
       * 48kHz playback may require 100MHz (6.25MHz x 16) operation and fast SD card

  Reference:
  -- https://ccrma.stanford.edu/courses/422/projects/WaveFormat/


  Note: This object requires four cogs to play a file (three when not playing)
  
}


con

  ' these pins are pre-defined for all Propeller apps

  RX     = 31                                                   ' programming port
  TX     = 30

  SDA    = 29                                                   ' i2c port
  SCL    = 28


con

  BUF_SIZE = 512                                                ' wav data buffer size    

  VOL_MAX  = 100
  VOL_MIN  =   0


con

  {
  
    Note on return values: File methods use 0 for okay/true, negative values
    for errors as does the underlying FSRW SD object
    
  }

  OKAY       =   0
  
  YES        =   0
  NO         =  -1

  NO_FILE    =  -1                                              ' file not found
  
  BAD_FILE   =  -2                                              ' does not match requirements
  BAD_HDR    =  -3
  BAD_FORMAT =  -4
  BAD_CHANS  =  -5
  BAD_RATE   =  -6
  BAD_BITS   =  -7
  
  NO_MOUNT   = -10

  NO_COG     = -20


  
obj

  dac : "stereo_dacs"                                           ' dithered dac driver (cog 1)
  sd  : "fsrw"                                                  ' file io
    

var

  long  lpin                                                    ' do not move/change 
  long  rpin
  long  sddo                                                    
  long  sdclk
  long  sddi
  long  sdcs
  long  dither
  
  long  lrsamples                                               ' for stereo_dacs

  long  dacscog                                                 ' cog running dacs driver (PASM)
  long  wavcog                                                  ' wav-to-dac cog (PASM)
  long  bufcog                                                  ' buffering cog (Spin)
  long  bcstack[64]                                             ' stack for buffering cog

  long  mounted
    
  long  playflag                                                ' playing
  long  filerate                                                ' original sample rate
  long  srate                                                   ' sample rate (ticks)
  long  buf0count                                               ' # samples in buf0
  long  buf0[BUF_SIZE]                                          ' raw samples
  long  buf1count                                              
  long  buf1[BUF_SIZE]
  
  long  lvolume                                                 ' volume settings (0 - 100%)
  long  rvolume
  
  long  speed                                                   ' playback, 50 to 100%

  byte  header[44]                                              ' for WAV header
  
  byte  wavcheck[5]

  

pub main

'' Dummy routine in case this file is accidentally loaded as top object

  repeat
    waitcnt(0)
 

pub start(p_pindefs) | check

'' Loads elements for wav file playback
'' -- returns true if SD card successfully mounted
'' -- returns false if SD card does not mount
''    * calls for play will attempt to re-mount SD card 

  stop

  longmove(@lpin, p_pindefs, 7)                                 ' get pins / setup
  dacscog := dac.start(@lrsamples, lpin, rpin, dither)          ' start stereo dac driver
  if (dacscog == 0)
    return NO_COG       

  ' set wav cog parameters
  ' -- write values into PASM before launching

  dacpntr   := @lrsamples           
  pfpntr    := @playflag    
  srpntr    := @srate   
  b0cntpntr := @buf0count
  b0pntr    := @buf0
  b1cntpntr := @buf1count 
  b1pntr    := @buf1 
  lvolpntr  := @lvolume
  rvolpntr  := @rvolume

  set_volume(MAX_VOL, MAX_VOL)                                  ' default to max
  set_speed(100)                                                ' default norm

  if (mount == OKAY)
    wavcog := cognew(@wavdacs, 0) + 1                           ' start wav driver cog
    return true
  else
    return false


pub stop

'' Stop audio object and unload any cogs in use

  \sd.unmount
  dac.stop
  
  if (wavcog)
    cogstop(wavcog - 1)
    wavcog := 0

  if (bufcog)
    cogstop(bufcog - 1)
    bufcog := 0

  bytefill(@header, 0, 44)                                      ' clear header

  
pri mount | check

'' Attempts to mount SD card
''   0 -- SD mounted
'' -10 -- SD not mounted

  sd.pclose                                                    ' close any open file

  check := \sd.mount_explicit(sddo, sdclk, sddi, sdcs)          ' attempt card mount
    
  if (check < OKAY)
    mounted := false
    return NO_MOUNT
  else
    mounted := true
    return OKAY


pub sd_mounted

'' Returns true if SD card is mounted

  return mounted

  
pub set_volume(l, r)

'' Set playback levels for left and right channels, 0 to 100%

  l := (VOL_MIN #> l <# VOL_MAX)                                ' keep in legal range
  if (l == 100)
    lvolume := $1_0000                                          ' avoid rounding errors
  else
    lvolume := l * $1_0000 / 100
   
  r := (VOL_MIN #> r <# VOL_MAX)
  if (r == 100)
    rvolume := $1_0000
  else
    rvolume := r * $1_0000 / 100


pub set_speed(spd)

'' Set playback speed, 50% to 200%
'' -- 48kHz files may requires 100MHz operation (6.25MHz x 16x PLL)

  speed := 50 #> spd <# 200
  srate := ((clkfreq / filerate) * 100 / speed) #> (clkfreq / 48_000) 


pub play(p_filename, n) | check

'' Starts file in p_filename, n times
'' -- 0 to loop forever
'' -- will stop player if active, even if bad file specifed

  stop_player                                                   ' stop if playing now

  if (!mounted)                                                 ' is card mounted?
    mount                                                       '  if no, attempt re-mout
    ifnot (mounted)                                             '   if fail
      return NO_MOUNT                                           '     exit with error code                                        

  check := check_file(p_filename, false)                        ' check file specs
  if (check == OKAY)                                            ' okay?
    bufcog := cognew(wav_spooler(p_filename, n), @bcstack) + 1  ' yes, start file
    if (bufcog == 0)
      return NO_COG
    else
      return OKAY
  else
    return BAD_FILE


pub playing

'' Returns player status

  return (bufcog > 0)

  
pub kill

'' Stops audio buffering cog if active

  if (playing)                                                   
    stop_player                                                 ' shutdown/clear buffers 
    sd.pclose                                                   ' close file             
    cogstop(bufcog - 1)                                         ' shutdown spooling cog
    bufcog := 0


pub has_file(p_filename)

'' Searches for file (at p_filename) on SD card

  stop_player                                                   ' stop if playing

  if (!mounted)                                                 ' is card mounted?
    mount                                                       '  if no, attempt re-mout
    if (!mounted)                                               '   if fail
      return NO_MOUNT                                           '     exit with error code                                        

  if (\sd.popen(p_filename, "r") == OKAY)                       ' try to open file
    return OKAY
  else
    return NO_FILE


pub get_header(p_filename)

'' Opens file at p_filename, moves header bytes into local buffer

  bytefill(@header, 0, 44)                                      ' clear buffer

  if (!mounted)                                                 ' is card mounted?
    mount                                                       '  if no, attempt re-mout
    if (!mounted)                                               '   if fail
      return NO_MOUNT                                           '     exit with error code  

  if (\sd.popen(p_filename, "r") <> OKAY)                       ' attempt to open
    sd.pclose
    return NO_FILE

  if (\sd.pread(@header, 44) == 44)                             ' get header bytes
    return OKAY
  else
    sd.pclose
    return BAD_HDR    
    

pub format | fmt

'' Returns format of WAV file
'' -- call get_header() before use

  fmt := 0                          
  bytemove(@fmt, @header[20], 2)

  return fmt
  

pub channels | ch

'' Returns channels in WAV file
'' -- call get_header() before use

  ch := 0                           
  bytemove(@ch, @header[22], 2)

  return ch


pub sample_rate  

'' Returns sample rate (Hz) of WAV file
'' -- call get_header() before use

  bytemove(@filerate, @header[24], 4)
  
  return filerate


pub bits_per_sample | bits

'' Returns bits per sample of WAV file
'' -- call get_header() before use

  bits := 0
  bytemove(@bits, @header[34], 2)

  return bits


pub samples | balign, chunk2

'' Returns samples in WAV file (chunk2)
'' -- call get_header() before use

  balign := 0
  bytemove(@balign, @header[32], 2)                             ' block align (bytes per sample)
  chunk2 := 0
  bytemove(@chunk2, @header[40], 4)                             ' bytes in chunk2 (audio)  

  return chunk2 / balign


pub audio_bytes | chunk2

'' Returns bytes in audio chunk
'' -- call get_header() before use

  bytemove(@chunk2, @header[40], 4)                             ' bytes in chunk2 (audio)  

  return chunk2


pub duration

'' Returns duration of audio chunk in 1/100ths seconds
'' -- call get_header() before use

  return (samples * 10) / (sample_rate / 10)


pub check_file(p_filename, remount) | check

'' Ensures file is present and correct type (PCM, mono or stereo, 16-bit)
'' -- must stop audio in progress to check file

  stop_player                                                   ' kill if running now

  if ((remount == true) or (!mounted))
    if (mount <> OKAY)
      return NO_MOUNT

  if (\sd.popen(p_filename, "r") <> OKAY)                        ' try to open  
    sd.pclose
    return NO_FILE

  check := \sd.pread(@header, 44)
  sd.pclose    
  if (check <> 44)
    return BAD_HDR                                              ' bad file size

  bytemove(@wavcheck, @header[0], 4)                            ' copy string
  wavcheck[4] := 0                                              ' terminate
  ifnot strcomp(@wavcheck, string("RIFF"))                      ' look for marker
    return BAD_FILE

  bytemove(@wavcheck, @header[8], 4)      
  wavcheck[4] := 0                        
  ifnot strcomp(@wavcheck, string("WAVE"))
    return BAD_FILE  

  bytemove(@wavcheck, @header[12], 3)      
  wavcheck[3] := 0                        
  ifnot strcomp(@wavcheck, string("fmt"))
    return BAD_FILE

  if (format <> 1)
    return BAD_FORMAT                                           ' not pcm
    
  if ((channels <> 1) and (channels <> 2))
    return BAD_CHANS                                            ' not mono or stereo

  if (sample_rate < 4000) or (sample_rate > 48_000)
    return BAD_RATE                                             ' bad sample rate

  if (bits_per_sample <> 16)
    return BAD_BITS                                             ' incorrect bit size

  bytemove(@wavcheck, @header[36], 4)      
  wavcheck[4] := 0                        
  ifnot strcomp(@wavcheck, string("data"))
    return BAD_FORMAT

  return OKAY                                                   ' no problems


pri stop_player

'' Stops player
'' -- clears play flag, rate, and play buffers

  playflag := false                                             ' stop playing
  srate := 0                                                    ' clear rate
  clearbufs                                                     ' clear buffers

  
pri start_player(rate) | tmp

'' Sets sample rate for driver and enables play flag
'' -- 48kHz files may requires 100MHz operation (6.25MHz x 16x PLL)

  if (speed == 100)
    srate := clkfreq / rate
  else
    srate := ((clkfreq / rate) * 100 / speed) #> (clkfreq / 48_000)
    
  playflag := channels                                          ' play now
 
  
pri wav_spooler(p_filename, plays) | n, playbytes, playcount    ' launch with cognew 

'' Spools WAV data to DACs engine      
'' -- this method must be launched into its own cog

  if (plays =< 0)
    plays := posx                                               ' if 0, play "forever"
    
  repeat plays
    sd.popen(p_filename, "r")                                   ' open the file
    sd.pread(@header, 44)                                       ' read the header
    start_player(sample_rate)                                   ' enable the player
    
    playbytes := audio_bytes                                    ' play only audio data
    playcount := 0 
         
    repeat 
      if (buf0count == 0)                                       ' if buf0 is empty
        n := sd.pread(@buf0, constant(BUF_SIZE * 4))            ' read from file
        if (n > 0)                                              ' have bytes?
          buf0count := n >> 2                                   ' yes, convert to samples
          playcount += n
          if (playcount => playbytes)
            quit
        else                                                    ' at end of file
          buf0count := 0                                                     
          buf1count := 0
          quit                                                  ' we're done
          
      if (buf1count == 0)
        n := sd.pread(@buf1, constant(BUF_SIZE * 4))
        if (n > 0) 
          buf1count := n >> 2
          playcount += n
          if (playcount => playbytes)
            quit
        else
          buf0count := 0 
          buf1count := 0
          quit

    repeat until ((buf0count == 0) and (buf1count == 0))        ' let buffers run out

    sd.pclose                                                   ' close file
    stop_player                                                 ' shutdown/clear buffers 

  stop_player   
  bufcog := 0                                                   ' alert foreground
  cogstop(cogid)                                                ' unload this cog

    
pri clearbufs

  longfill(@buf0count, 0, (1 + BUF_SIZE) << 1)                  ' clear audio buffers
  bytefill(@header, 0, 44)                                      ' remove old header data


con

  { ===================================== }
  {                                       }
  {  B A S I C   F S R W   M E T H O D S  }
  {                                       }
  { ===================================== }


pub popen(p_filename, cmd)

  return \sd.popen(p_filename, cmd)

  
pub pclose

  return sd.pclose


pub get_filesize

  return \sd.get_filesize


pub pread(p_buf, count)

  return \sd.pread(p_buf, count)


pub pgetc

  return \sd.pgetc


pub pwrite(p_buf, count)

  return \sd.pwrite(p_buf, count)


pub pputs(p_zstr)

  return \sd.pputs(p_zstr)


pub hasfile(p_filename) | check

'' Returns true if file on SD card

  check := \sd.popen(p_filename, "r")
  if (check == 0)
    return true
  else
    return false    

  
pub readln(p_buf, n) | c, len

'' Reads line of text from open file into buffer
'' -- terminated by EOF or LF
'' -- removes linefeeds if present

  len := 0                                                     ' index into string

  repeat
    c := \sd.pgetc                                              ' get a character
    if (c < 0)                                                  ' end of file
      if (len == 0)                                             ' if nothing
        return -1                                               ' return empty
      else
        quit

    if (c <> 10)                                                ' if not LF char
      byte[p_buf++] := c                                        '  add to buffer
      len++
    else                                                        ' if LF
      quit                                                      '  we're done

  byte[p_buf] := 0                                              ' terminate end of line

  return len


con

  { ============================= }
  {                               }
  {  W A V D A C S   D R I V E R  }
  {                               }
  { ============================= }
 

dat

{

  Based in part on code by Ray Allen

}

                        org     0

wavdacs
quiet                   wrlong  ZERO, dacpntr                   ' write silence to StereoDuty
                        
                        rdlong  mode, pfpntr            wz      ' read play flag
        if_z            jmp     #quiet
        
                        rdlong  ratetix, srpntr         wz      ' sample rate tix
        if_z            jmp     #quiet
                        
                        call    #check0                         ' anything in 1st buffer?
        if_z            jmp     #quiet

                        ' ready to play

                        mov     smpltimer, cnt
                        add     smpltimer, ratetix

playloop                rdlong  mode, pfpntr            wz      ' check play flag
        if_z            jmp     #quiet

                        test    mode, #%10              wc      ' test channels
        if_c            jmp     #playstereo

        

' **********************
' M O N O
' **********************
'
playmono                waitcnt smpltimer, ratetix              ' let sample rate timer expire  

                        rdlong  ratetix, srpntr         wz      ' get sample rate tix
        if_z            jmp     #quiet
     
                        rdlong  xsamples, bufpntr               ' get two mono samples
                        mov     lchan, xsamples                 ' copy to left
                        mov     rchan, xsamples                 ' copy to right 
                        rdlong  lvol, lvolpntr                  ' get left volume setting
                        and     lchan, hx0000_FFFF              ' low sample 1st  
                        and     rchan, hx0000_FFFF              ' low sample 1st   
                        rdlong  rvol, rvolpntr                  ' get right volume setting  
                        
                        call    #playit  

                        waitcnt smpltimer, ratetix              ' let sample timer expire  

                        mov     lchan, xsamples                 ' copy 
                        mov     rchan, xsamples 
                        shr     lchan, #16                      ' extract 2nd samples
                        shr     rchan, #16   
                          
                        call    #playit  
                        jmp     #updatebuffer
        

                        
' **********************
' S T E R E O
' **********************
'
playstereo              waitcnt smpltimer, ratetix              ' let sample rate timer expire

                        rdlong  ratetix, srpntr         wz      ' sample rate tix
        if_z            jmp     #quiet

                        rdlong  xsamples, bufpntr               ' read samples (l + r)
                        mov     lchan, xsamples                 ' copy to left
                        mov     rchan, xsamples                 ' copy to right  
                        rdlong  lvol, lvolpntr                  ' get left volume    
                        and     lchan, hx0000_FFFF              ' extract left sample   
                        shr     rchan, #16                      ' extract right sample 
                        rdlong  rvol, rvolpntr                  ' get right volume

                        call    #playit



' update current buffer
' -- if done, zero hub count and check for data in next buffer
                        
updatebuffer            add     bufpntr, #4                     ' point to next sample
                        djnz    bcount, #playloop               ' done with this buffer?

                        wrlong  ZERO, bcpntr                    ' clear this buffer for re-fill
                        test    bufid, #1               wc      ' which buffer just played?
                        
        if_c            call    #check0                         ' anything in buffer 0?
        if_nc           call    #check1                         ' anything in buffer 1?
        if_nz           jmp     #playloop                       ' we're still playing

                        wrlong  ZERO, pfpntr                    ' clear play flag
                        jmp     #quiet                  


                        
' preps samples and sends to DACs driver
' -- lchan and rchan hold raw samples
' -- lvol and rvol hold volume levels

playit                  mov     arg1, lchan                     ' adjust left channel
                        mov     arg2, lvol 
                        call    #adjust
                        mov     lchan, arg1

                        mov     arg1, rchan                     ' adjust right channel
                        mov     arg2, rvol 
                        call    #adjust
                        mov     rchan, arg1

                        shl     rchan, #16                      ' move right to high word  
                        or      lchan, rchan                    ' reconstruct        
                        wrlong  lchan, dacpntr                  ' update dac driver 
playit_ret              ret



' Multiply signed, 16-bit channel data by unsigned (16-bit) adjustment
' -- adjustment expressed as percent (0.00 to 1.00) x $1_0000
'
' arg1 = channel value in (signed, 16-bit)
' arg2 = adjustment (unsigned)
' t1 = holds sign bit
' t2 = working variable

adjust                  cmp     arg2, hx0000_FFFF       wc      ' 100%?
        if_a            jmp     #adjust_ret

                        cmps    arg2, #1                wc      ' 0%
        if_b            mov     arg1, #0
        if_b            jmp     #adjust_ret

                        shl     arg1, #16                       ' extend sign from bit15
                        sar     arg1, #16
                        abs     arg1, arg1              wc      ' get sign, force positive
                        muxc    t1, #1                          ' save sign

                        max     arg1, POSX16                    ' limit to 16-bit pos max
                        mov     t2, #0                          ' clear result workspace                   
                                             
:loop                   shr     arg2, #1                wc      ' get multiplier bit
        if_c            add     t2, arg1                        ' if 1, add muliplicand
                        shl     arg1, #1                wc      ' adjust for next bit
                        tjnz    arg2, #:loop                    ' again if more bits
                        
                        test    t1, #1                  wc      ' recall and restore sign
                        negc    arg1, t2
                        shr     arg1, #16                       ' back to 16 bits
adjust_ret              ret



' check samples in buffer
' -- resets buffer id and pointer to buffer
' -- sets Z flag if buffer is empty

check0                  mov     bufid, #0                       ' set buffer id
                        mov     bcpntr, b0cntpntr               ' point to count
                        rdlong  bcount, bcpntr          wz      ' get buffer samples count
                        mov     bufpntr, b0pntr                 ' point to element 0
check0_ret              ret


check1                  mov     bufid, #1
                        mov     bcpntr, b1cntpntr      
                        rdlong  bcount, bcpntr          wz         
                        mov     bufpntr, b1pntr         
check1_ret              ret

' -------------------------------------------------------------------------------------------------                        

ZERO                    long    0
MAX_VOL                 long    $1_0000
POSX16                  long    $0_7FFF

hx0000_FFFF             long    $0000_FFFF                      ' for truncating to 16 bits
hx8000_0000             long    $8000_0000                      ' zero level out at 32 bits

dacpntr                 long    0-0                             ' samples address for StereoDuty

pfpntr                  long    0-0                             ' pointer to playflag
srpntr                  long    0-0                             ' pointer to sample rate ticks
b0cntpntr               long    0-0                             
b0pntr                  long    0-0                             
b1cntpntr               long    0-0                             
b1pntr                  long    0-0
lvolpntr                long    0-0
rvolpntr                long    0-0

mode                    res     1                               ' mono or stereo

ratetix                 res     1                               ' ticks in sample rate
smpltimer               res     1                               ' timer for samples
bufid                   res     1                               ' buffer in use
bcpntr                  res     1                               ' buffer count pointer
bcount                  res     1                               ' samples in buffer
bufpntr                 res     1                               ' pointer to buffer in use
lchan                   res     1                               ' left ch sample
rchan                   res     1                               ' right ch sample
lvol                    res     1                               ' left volume
rvol                    res     1                               ' right volume

xsamples                res     1                               ' samples transferred from hub

arg1                    res     1                               ' arguement
arg2                    res     1                               ' arguement

t1                      res     1
t2                      res     1

                        fit     496


dat

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}